{
Dumps a floppy diskette to an image file.  Written by trixter@oldskool.org

This imager has special handling for disks with bad sectors:
- It will try to re-read all bad sectors after the good tracks have been saved
- If multiple re-reads fail, the sector will be dumped as all "!!!!!"s to
  clearly indicate which sectors could not be read

The behavior of dumping "!!!!"s for unreadable sectors is known as the "Better
the devil you know, than the devil you don't" strategy.
}

{{$DEFINE DEBUG}

Program disk2img;

uses
  flprglob,   {Flopper utils unit; global variables}
  flprdisk,   {Flopper utils unit; low-level disk utilities}
  flprsprt,   {Flopper utils unit; additional support routines}
  support,    {misc support routines}
  TInterrupts,
  m6845ctl,   {cgalib}
  cmdlin;     {external "getopt"-like library by Jim Walsh, CIS:72571,173}

const
  badsec9s:byte=0;  {quick'n'dirty format checking}
  badhead1s:word=0; {quick'n'dirty format checking}
  recoverCount:word=0;
  refineCount:word=0;
  refineLimit=5;    {if reconst. sector doesn't change this times, finish up}
  maxRecoverCount:word=1024;
  Recover:boolean=false;
  Reconstruct:boolean=false;
  Query:boolean=false;
  TrackDump:boolean=true;
  RenameExt:boolean=false;
  ignoreErrors:boolean=false;
  salvageSecLine=12;

type
  phase=(initial,recovery);

var
  outf:file;
  filename:string;
  readStatus:array[0..maxTracks-1,0..maxHeads-1,1..maxSectors] of boolean;

{------Main program functions------------------------------------------------}

Procedure DumpInit;

const
  defaultName='output.img';

  Procedure printhelp;
  begin
    writeln('Usage is:  "DISK2IMG d: filename.ext [options]", where "d:" is drive letter.');
    writeln;
    writeln('Option  Description                                           Default');
{    writeln('~~~~~~  ~~~~~~~~~~~                                          ~~~~~~~');}
    writeln('  /Txx  number of Tracks                                      /t',numtracks);
    writeln('  /Sxx  number of Sectors per track                           /s',numsectors);
    writeln('  /Hx   number of Heads                                       /h',numheads);
    writeln('  /R    Reconstruct bad sectors (CAUTION! SEE DOCUMENTATION)  ',Reconstruct);
    writeln('  /Axxx Number of sector read attempts                        ',maxRecoverCount);
    writeln('  /E    Rename file extension to match image size             ',renameExt);
    writeln('  /Q    Query boot sector to get disk parameters              ',Query);
    writeln('  /I    Ignore errors (only useful for non-DOS disks)         ',ignoreErrors);
    (*writeln('  /Dx   optional way to specify drive to work on              /d',upcase(chr(65+CurDrive)));*)
    writeln;
    writeln('Default settings reflect a ',numtracks*numsectors,'K diskette in drive ',upcase(chr(65+CurDrive)),':');
    writeln('If no filename is specified, the name "'+defaultname+'" will be used.');
    writeln;
    writeln('Examples:');
    writeln('DISK2IMG a: mygame.img                         dumps disk in A: to "mygame.img"');
    writeln('DISK2IMG b: bootdisk.img /q /e    query disk geometry and rewrite .ext to match');
    halt(1);
  end;

  Procedure GetOptions;
  var
    foos:string;
    fooc:char;

  begin
    {help}
    if is_param('?') or (non_flag_count=0) then printhelp;

    {new filename?}
    filename:=defaultName;
    if non_flag_count>0 then begin
      filename:=non_flag_param(1);
      {if filename is "A:", "B:" etc. then pick the other string as filename}
      if filename[2]=':'
        then if non_flag_count>1
          then filename:=non_flag_param(2)
          else filename:=defaultName;
    end;

    {drive to work on}
    (*if is_param('d') then begin
      foos:=param_text('d');
      CurDrive:=ord(upcase(foos[1]))-65;
    end;*)
    {Also handle if first string is less than 3 chars since that is a drive}
    if (non_flag_count<>0) and (length(non_flag_param(1))<3) then begin
      foos:=non_flag_param(1);
      CurDrive:=ord(upcase(foos[1]))-65;
    end;

    {number of tracks}
    if is_param('t') then numtracks:=param_int('t');

    {number of sectors per track}
    if is_param('s') then numsectors:=param_int('s');

    {number of heads}
    if is_param('h') then numheads:=param_int('h');

    {number of recovery attempts}
    if is_param('a') then maxRecoverCount:=param_int('a');
    if maxRecoverCount>(60*300)
      then fatalerror(99,inttostr(maxRecoverCount)+' attempts takes '
           +inttostr(maxRecoverCount div 300)+' minutes per sector and will shred your disk!');

    {query disk boot record for track, heads, spt}
    if is_param('q') then Query:=true;

    {Rename filename extension to match size of image?}
    if is_param('e') then renameExt:=true;

    {Ignore drive reset errors -- good for trying to recover bad disks}
    if is_param('i') then ignoreErrors:=true;

    {Attempt to salvage bad sectors by reconstructing them}
    if is_param('r') then Reconstruct:=true;
  end;

  Procedure SetupVars;
  var
    status:word;
    tmpe:string[12];
    dsp:^boot_structure;
    _t,_h,_s:byte;

  begin
    {Init sector status buffer}
    {$IFDEF DEBUG}
    fillchar(readStatus,sizeof(readstatus),byte(true));
    for _t:=20 to numTracks-1 do
      for _h:=0 to numHeads-1 do
        for _s:=1 to numSectors do
          readStatus[_t,_h,_s]:=false;
    {$ELSE}
    fillchar(readStatus,sizeof(readstatus),byte(false));
    {$ENDIF}
    {this should be the ONLY time we reset the drive subsystem}
    status:=ResetDrive(CurDrive);
    if (status<>0) and not ignoreErrors
      then fatalerror(1,'"'+DriveStatusString(status)+'" when resetting drive subsystem');

    if Query then begin
      getmem(sectorsbuffer,sectorsize);
      status:=ReadSectors(CurDrive, 0, 0, 1, 1, sectorsbuffer);
      if ((status and $00FF)<>1) {should get number of sectors read as return value}
        then fatalerror(2,'Nothing read; "'+DriveStatusString(GetDriveStatus(curdrive))+'" when trying to query boot sector');
      dsp:=sectorsbuffer;
      if (dsp^.sectors_per_track = 0)
      or (dsp^.sectors_per_track > 18)
      or (dsp^.number_of_heads = 0)
      or (dsp^.number_of_heads > 2)
        then fatalerror(3,'Not a DOS-formatted diskette; re-run without /Q');
      numsectors:=dsp^.sectors_per_track;
      numheads:=dsp^.number_of_heads;
      numtracks:=dsp^.total_sectors div numsectors div numheads;
      freemem(sectorsbuffer,sectorsize);
    end;

    tracksize:=sectorsize*numsectors*numheads;
    totalsectors:=numtracks*numsectors*numheads;
    totaltrackreads:=numtracks*numheads;
    getmemDMAPageAligned(sectorsbuffer,tracksize);

    if renameExt then begin
      status:=pos('.',filename); {grab location of period}
      if status<>0 then begin
        delete(filename,status+1,length(filename)-status);
        tmpe:=inttostr(((numtracks*numsectors) div 2)*numheads); {make extension "360" or "1440" etc.}
        if length(tmpe) > 3 then byte(tmpe[0]):=3; {truncate extension if necessary}
        filename:=filename+tmpe;
      end;
    end;

    salvagedSectorp:=@salvagedSector;
    oldSectorp:=@oldSector;
  end;

  Procedure PrepDrive;
  var
    chkstatus:word;
    sec:byte;
  begin
    {initialize the floppy controller subsystem}
    {read dummy sector three times in a row to force drive to spin up}
    DisplayStatus('Forcing motor to spin...');
    for sec:=0 to 2 do begin
      DisplayStatus('Reading sector #0...');
      {read only first sector since that is the only sector guaranteed to be readable}
      chkstatus:=ReadSectors(CurDrive,0,0,1,1,sectorsbuffer);
      if lo(chkstatus)<>1 {no sectors read} then begin
        DisplayStatus('Wrong number of sectors read; resetting drive...');
        resetdrive(CurDrive);
      end else break;
    end;

    {if we never read anything successfully, abort}
    if (lo(chkstatus)<>1) and not ignoreErrors
      then fatalerror(10,'All sample reads returned errors; reboot to reset the BIOS disk subsystem.');

    DisplayStatus('Checking drive motor...');
    if (DriveMotorStatus and $F = 0) and not recover
      then fatalerror(5,'Disk drive did not spin up!');
  end;

begin
  filename:=defaultname;
  PrintWelcome;
  GetOptions;
  SetupVars;
  SetupScreen;
  PrepDrive;
end;

Procedure DumpDo(mode:phase);

var
  curstatus:word;
  ch:char;
  sector,head,track:byte;
  destp:^byte;
  status:byte;
  tr,se,he:byte;
  tmps:string;
  abspos:longint;
  dataPending:boolean;

  Procedure chkForUserAbort;
  var
    ch:char;
  begin
    if keypressed then begin
      ch:=readkeychar;
      if ch=#27 then begin
        close(outf);
        fatalerror(0,'User aborted');
      end;
    end;
  end;

  Procedure reconstructSector;assembler;
  {
  The idea is to write the sector out using the average of each read bit.
  For each original bit position in the sector, the total bits read is
  divided by the # of reads, rounded up (meaning, 0.5 or higher is 1,
  otherwise 0).  salvagedSector and salvageData are both in DS.
  salvageData[(512*8)*2] holds the # of reads.
  }
  asm
@start:
        mov     si,offset salvageData
        mov     di,offset salvagedSector
        mov     bp,word ptr salvageData[(sectorsize*8)*2] {ax:=count / total}
        mov     cx,sectorsize
        push    ds
        pop     es
        cld
@bytesloop:
        xor     bx,bx                   {reconstructed byte}

        {unrolled 8 times, once for each bit, so we can gain a register}
        lodsw                           {grab a bit count (word)}
        mov     dx,ax
        shr     dx,1
        add     ax,dx                   {ax:=count + (count/2), for rounding}
        xor     dx,dx
        div     bp                      {ax:=count / total}
        shr     ax,1                    {get average rounded bit into carry}
        rcl     bl,1                    {rotate carry into reconstructed byte}

        lodsw                           {grab a bit count (word)}
        mov     dx,ax
        shr     dx,1
        add     ax,dx                   {ax:=count + (count/2), for rounding}
        xor     dx,dx
        div     bp                      {ax:=count / total}
        shr     ax,1                    {get average rounded bit into carry}
        rcl     bl,1                    {rotate carry into reconstructed byte}

        lodsw                           {grab a bit count (word)}
        mov     dx,ax
        shr     dx,1
        add     ax,dx                   {ax:=count + (count/2), for rounding}
        xor     dx,dx
        div     bp                      {ax:=count / total}
        shr     ax,1                    {get average rounded bit into carry}
        rcl     bl,1                    {rotate carry into reconstructed byte}

        lodsw                           {grab a bit count (word)}
        mov     dx,ax
        shr     dx,1
        add     ax,dx                   {ax:=count + (count/2), for rounding}
        xor     dx,dx
        div     bp                      {ax:=count / total}
        shr     ax,1                    {get average rounded bit into carry}
        rcl     bl,1                    {rotate carry into reconstructed byte}

        lodsw                           {grab a bit count (word)}
        mov     dx,ax
        shr     dx,1
        add     ax,dx                   {ax:=count + (count/2), for rounding}
        xor     dx,dx
        div     bp                      {ax:=count / total}
        shr     ax,1                    {get average rounded bit into carry}
        rcl     bl,1                    {rotate carry into reconstructed byte}

        lodsw                           {grab a bit count (word)}
        mov     dx,ax
        shr     dx,1
        add     ax,dx                   {ax:=count + (count/2), for rounding}
        xor     dx,dx
        div     bp                      {ax:=count / total}
        shr     ax,1                    {get average rounded bit into carry}
        rcl     bl,1                    {rotate carry into reconstructed byte}

        lodsw                           {grab a bit count (word)}
        mov     dx,ax
        shr     dx,1
        add     ax,dx                   {ax:=count + (count/2), for rounding}
        xor     dx,dx
        div     bp                      {ax:=count / total}
        shr     ax,1                    {get average rounded bit into carry}
        rcl     bl,1                    {rotate carry into reconstructed byte}

        lodsw                           {grab a bit count (word)}
        mov     dx,ax
        shr     dx,1
        add     ax,dx                   {ax:=count + (count/2), for rounding}
        xor     dx,dx
        div     bp                      {ax:=count / total}
        shr     ax,1                    {get average rounded bit into carry}
        rcl     bl,1                    {rotate carry into reconstructed byte}

        xchg    bx,ax
        stosb                           {store byte to sector buffer}

        loop    @bytesloop              {do all bytes in the sector}
  end;

begin
  curstatus:=0;
  numErrors:=0;
  assign(outf,filename);

  {If initial run, we dump all tracks quickly.  If a second run is needed
  because we got errors, we attempt to dump only sectors that had problems.}

  if mode=initial then begin
    rewrite(outf,1);
    trackDump:=true;
    Recover:=false;
   end else begin
    reset(outf,1);
    trackDump:=false;
    Recover:=true;
  end;

  {calc actual # of sectors left based on read matrix; used for % time left}
  starttick:=TicksSinceMidnight;
  sectorcount:=0;
  totalsectors:=0;
  for track:=0 to numtracks-1 do
    for head:=0 to numheads-1 do
      for sector:=1 to numsectors do
        if readStatus[track,head,sector]=false
          then inc(totalsectors);

  {$IFDEF DEBUG}
  for track:=20 to numtracks-1 do begin
  {$ELSE}
  for track:=0 to numtracks-1 do begin
  {$ENDIF}
    {reset destintation to start of track buffer}
    destp:=sectorsbuffer;
    dataPending:=false;
    for head:=0 to numheads-1 do
      {fill a track buffer (both heads) before writing to file and moving on}
      if trackdump then begin
        sector:=1; {we're reading an entire track so start at sector 1}
        {status returned in AH, num sectors read in AL}
        DisplayDriveStatus(track,head,0,false); {0=doing a track dump}
        curstatus:=ReadSectors(CurDrive,head,track,sector,numsectors,destp);
        dataPending:=true;
        DisplaySector(destp,0);
        tmps:='Last status code: '+DriveStatusString(hi(curstatus));
        DisplayStatus(tmps);

        if hi(curstatus)<>0 then begin
          inc(numErrors);
        end else begin
          {mark sectors as read successfully}
          for se:=1 to numSectors do
            readStatus[track,head,se]:=true;
        end;

        {increment dest. pointer if we have more heads to read}
        inc(longint(destp),tracksize div numheads);
        inc(trackreadcount);
        chkForUserAbort;
      end else begin
        {second phase, recover bad sectors}
        for sector:=1 to numsectors do begin
          {Did we read this sector successfully already?  Skip if so}
          if readStatus[track,head,sector]=true then continue;
          DisplayDriveStatus(track,head,sector,false);
          curstatus:=ReadSectors(CurDrive,head,track,sector,1,destp);
          DisplayStatus('Last status code: '+DriveStatusString(hi(curstatus)));
          dataPending:=true;
          recoverCount:=0;
          if Reconstruct then fillchar(salvageData,sizeof(salvageData),0);
          if Recover and (lo(curstatus)<>1) then repeat
            inc(recoverCount);
            DisplayDriveStatus(track,head,sector,true);
            {read directly into track buffer on the heap}
            curstatus:=ReadSectors(CurDrive,head,track,sector,1,destp);
            {was this read good?  If so, stop retrying the read}
            if lo(curstatus)=1 then continue;
            if Reconstruct then begin
              if recoverCount=1 then DisplayStatus('Attempting reconstruction');
              {copy into reconstruction buffer}
              move(destp^,salvagedSector,sectorsize);
              {update reconstruction iteration count}
              salvageData[sectorsize*8]:=recoverCount;
              {update reconstruction stats}
              updateSalvageCounts;
              DisplaySector(salvagedSectorp,0);
            end else begin
              DisplayStatus('(Recover) Drive status: '+DriveStatusString(hi(curstatus))
              +'  Re-Reading, attempt #'+inttostr(recoverCount));
              DisplaySector(destp,0);
            end;
            {Reconstruction is an expensive operation at 4.77 MHz,
            and we also want the heads off the media as soon as possible...
            but it is neat to watch it refine real-time, so we'll
            reconstruct the sector and user display every 4th read.}
            if (Reconstruct=true) and ((recoverCount AND 3)=0) then begin
              reconstructSector;
              DisplayStatus('Reconstructed sector from '+inttostr(recoverCount)+' samples...');
              DisplaySector(salvagedSectorp,salvageSecLine);
              {if the rebuilt sector stops changing, we can early abort}
              if BuffersEqual(salvagedSectorp,oldSectorp,sectorsize)<>0 then begin
                {copy current reconstructed buffer to storage}
                move(salvagedSector,oldSector,sizeof(oldSector));
                refineCount:=0;
              end else begin
                {they're equal -- make a note of it}
                inc(refineCount);
                DisplayStatus('Reconstructed sector appears stable; concluding refinement in '
                  +inttostr(refineLimit-refineCount)+' iterations');
                {if they've been equal for a threshold, conclude reconstruction}
                if refineCount>=refineLimit
                  then break;
              end;
            end;
            {handle user input clumsily here}
            if (keypressed=true) then begin
              {empty the keyboard buffer}
              if keypressed then while keypressed do
                ch:=upcase(readkeychar);
              {if key was R then recalibrate the track,
              else assume user wants to abort this sector}
              if ch='R' then begin
                ch:=#0;
                status:=ResetDrive(CurDrive);
                DisplayStatus('Recalibrating drive and track...')
              end else begin
                if Reconstruct
                  then DisplayStatus('(Reconstruction) Finishing up...')
                  else DisplayStatus('(Recover) Discarding this sector...');
                break;
              end;
            end;
          until (lo(curStatus)=1) or (recoverCount>maxRecoverCount); {until 1 sector read or out of time}

          if lo(curstatus)<>1 then begin
            {If we're here, we didn't read a sector successfully.}
            {Record some notes for quick'n'dirty error checking}
            if (sector=9) then inc(badsec9s);
            if (head=(numheads-1)) then inc(badhead1s);
            {mark the sector as failed}
            readStatus[track,head,sector]:=false;
            {either invalidate the sector, or write the reconstruction}
            if Reconstruct then begin
              {perform final reconstruction}
              reconstructSector;
              move(salvagedSector,destp^,sectorsize);
            end else begin
              {fill buffer with discard value}
              fillchar(destp^,sectorsize,'!');
            end;
          end else begin
            {We got a good read!  Update the status matrix}
            {DisplayStatus('Read successful!');}
            readStatus[track,head,sector]:=true;
          end;
          {Display whatever we read, reconstructed, or discarded}
          DisplaySector(destp,0);
          {increment the track buffer offset for next time}
          inc(longint(destp),sectorsize);
          inc(sectorcount); {used for estimated time display}
          chkForUserAbort;
        end;
      end;
    {If we have new track data, patch it into the image file}
    if dataPending then begin
      abspos:=tracksize;
      abspos:=abspos * track; {stupid arithmetic overflow workaround}
      seek(outf,abspos);
      blockwrite(outf,sectorsBuffer^,TrackSize);
    end;
  end;
  close(outf);
end;

Procedure DumpDone;

  Procedure FreeVars;
  begin
    freemem(sectorsbuffer,sectorsize);
  end;

  Procedure CleanUp;

    procedure ReportError(actualtype,newswitch:string);
    begin
      writeln('The disk you just dumped appears to ',actualtype,'.');
      writeln('You should re-dump it with the "-',newswitch,'" command-line switch.');
      writeln;
    end;

  var
    l:longint;
    _t,_h,_s:byte;

  begin
    doneScreen;

    writeln('Done.');
    {any "AI" to report?}
    {if number of bad sector 9's == the number of tracks*heads, then
     every single sector 9 didn't read--which means it's probably an 8-sector
     disk.  Warn user appropriately.}
    if badsec9s=(numtracks*numheads)
      then reporterror('have 8 sectors per track','s8');
    {if number of bad sectors on the second side == everything then
     every single sector on the second side didn't read--which means it's probably a
     single-sided disk.  Warn user appropriately.}
    if badhead1s=(numtracks*numsectors)
      then reporterror('be a single-sided disk','h1');
    if numErrors<>0 then begin
      Chan2SquarewaveOn(200);
      writeln(numErrors,' errors dumping this disk; consider re-dumping/deleting the image.');
      l:=ticksSinceMidnight+18; repeat until ticksSinceMidnight > l;
      Chan2SquarewaveOff;
    end;
    writeln('');

    for _t:=0 to numTracks-1 do
      for _h:=0 to numHeads-1 do
        for _s:=1 to numSectors do
          if readStatus[_t,_h,_s]=false then begin
            write('Track: ',_t,' Head: ',_h,' Sector: ',_s,' unreadable');
            if Reconstruct
              then writeln('; reconstruction was attempted')
              else writeln('; dumped as '+inttostr(sectorsize)+ ' "!" bytes instead');
          end;

  end;

begin
  FreeVars;
  CleanUp;
end;

begin
  DumpInit;
  DumpDo(initial);
  if numerrors>0
    then DumpDo(recovery);
  DumpDone;
end.
