unit flprdisk;

interface

FUNCTION ReadSectors(drive, head, track, sector, numtoread : Byte; buff: Pointer): word;
FUNCTION WriteSector(drive, head, track, sector : Byte; buff: Pointer): word;
FUNCTION ResetDrive(drive:byte): Byte;
FUNCTION GetDriveStatus(drive:byte):byte;
Procedure getmemDMAPageAligned(var DMAp:pointer;size:word);

var
  DriveMotorStatus:Byte Absolute $40:$3F;

implementation

uses
  support,
  flprglob;

Procedure getmemDMAPageAligned(var DMAp:pointer;size:word);
var
  basep,finalp:pointer;
  b:byte;
  w:word;
begin
  {allocate the buffer and see if it's within a DMA page boundary}
  getmem(basep,size);
  finalp:=l2p(p2l(basep)+size);
  asm
        mov     al,byte ptr basep+3
        mov     bl,byte ptr finalp+3
        mov     cl,4
        shr     al,cl
        shr     bl,cl
        sub     bl,al
        mov     b,bl
  end;
  if b<>0 then begin
    {The buffer crosses a DMA page boundary, so we will unallocate our
    previous buffer, then recreate it ending at the page boundary, then
    create a new buffer right on the boundary.}
    asm
        and     word ptr finalp+2,0F000h
        mov     word ptr finalp,0000h   {finalp is now our ideal}
    end;
    w:=p2l(finalp)-p2l(basep); {difference between existing and ideal}
    freemem(basep,size); {basep:=nil; finalp:=nil;}
    getmem(basep,w); {allocate new spacer buffer ending at page boundary}
    getmem(finalp,size); {allocate final buffer starting at page boundary}
    freemem(basep,w); {return spacer buffer to heap}
    DMAp:=finalp;
  end else begin
    DMAp:=basep;
  end;
end;

FUNCTION ReadSectors(drive, head, track, sector, numtoread : Byte; buff: Pointer): word; assembler;
{ drive = 0 for drive A:, 1 = B:,   }
{ 80h = first hard drive.           }
ASM
  mov     ah,02h
  mov     al,numtoread
  les     bx,buff     { es:bx -> buffer }
  mov     ch,track
  mov     cl,sector
  mov     dh,head
  mov     dl,drive
  int     13h
  {numsectors read in al, status in ah, cf=0 if successful}
END;

FUNCTION WriteSector(drive, head, track, sector : Byte; buff: Pointer): word; assembler;
{ drive = 0 for drive A:, 1 = B:,
{ 80h = first hard drive. }
ASM
  mov     ax,0301h      { read 1 sector }
  les     bx,buff       { es:bx -> buffer }
  mov     ch,track
  mov     cl,sector
  mov     dh,head
  mov     dl,drive
  int     13h
  mov     al,ah         { status result in al }
  pop     ds
END;

FUNCTION ResetDrive(drive:byte): Byte; assembler;
{ drive = 0 for drive A:, 1 = B:,   }
{ 80h = first hard drive.           }
ASM
  xor     ax,ax
  mov     dl,drive
  int     13h
  mov     al,ah         { return status result in al }
END;

FUNCTION GetDriveStatus(drive:byte): Byte; assembler;
ASM
  mov     ah,1
  int     13h
  { return status result in al }
END;


end.