unit flprsprt;
{User-facing I/O support routines (screen, messages, beeps, etc.)}

interface

Function upstring(name:string):string;
Procedure blip(freq,duration:word);
Procedure PrintHeader(progname,progdesc,progvers:string;progdate:longint);
Procedure PositionCursor(x,y:byte);
Function DriveStatusString(status:byte):string;
Procedure PrintWelcome;
Procedure DisplaySector(secbuf:pointer;line:byte);
function secstomins(secs:real):string;
Procedure DisplayStatus(s:string);
Procedure DisplayDriveStatus(_track,_head,_sector:byte;timeonly:boolean);
Procedure SetupScreen;
Procedure DoneScreen;
Procedure updateSalvageCounts;

implementation

uses
  m6845ctl,
  flprglob,
  support,
  TFastWrite, {fast textmode writing routines}
  TInterrupts;

Procedure blip(freq,duration:word);
var
  loops:word;
  l:longint;
begin
  {sound}
  Chan2SquarewaveOn(freq);
  {delay}
  l:=tickssincemidnight;
  for loops:=(duration div 18) downto 0 do begin
    repeat until l<>tickssincemidnight;
    l:=tickssincemidnight;
  end;
  {nosound}
  Chan2SquarewaveOff;
end;

Function upstring(name:string):string;
var
  foo:string;
  loop:byte;
begin
  for loop := 1 to Length(name) do
    foo[loop] := UpCase(name[loop]);
  foo[0]:=name[0];
  upstring:=foo;
end;

Procedure PrintHeader(progname,progdesc,progvers:string;progdate:longint);
const
  spacer='  ';
{prints a standard welcome header for the utilities}
begin
  Writeln(progname,spacer,progdesc,spacer,'Version ',progvers,spacer,progdate);
  {Writeln('Written by trixter@oldskool.org for the (now defunct) Flopper Project.');}
  {writeln('Use /? for available command-line options.');}
end;

Function DriveStatusString(status:byte):string;
begin
  case status of
    $00:DriveStatusString:='00: status OK';
    $01:DriveStatusString:='01: invalid function in AH or invalid parameter';
    $02:DriveStatusString:='02: address mark not found';
    $03:DriveStatusString:='03: disk write-protected';
    $04:DriveStatusString:='04: sector not found';
    $06:DriveStatusString:='06: floppy disk removed';
    $08:DriveStatusString:='08: DMA overrun';
    $09:DriveStatusString:='09: attempted DMA across 64K boundary';
    $0C:DriveStatusString:='0C: bad media type or invalid media';
    $10:DriveStatusString:='10: uncorrectable CRC error';
    $20:DriveStatusString:='20: controller failed';
    $40:DriveStatusString:='40: seek failed';
    $80:DriveStatusString:='80: timeout (disk not ready)';
    { hard-drive and special codes, for anyone interested }
    $05:DriveStatusString:='05: reset failed (hard disk)';
    $07:DriveStatusString:='07: drive parameter activity failed (hard disk)';
    $0A:DriveStatusString:='0A: bad sector detected (hard disk)';
    $0B:DriveStatusString:='0B: bad track detected (hard disk)';
    $0D:DriveStatusString:='0D: invalid number of sectors on format (PS/2 hard disk)';
    $0E:DriveStatusString:='0E: control data address mark detected (hard disk)';
    $0F:DriveStatusString:='0F: DMA arbitration level out of range (hard disk)';
    $10:DriveStatusString:='10: uncorrectable CRC or ECC data error!';
    $11:DriveStatusString:='11: data ECC corrected (hard disk)';
    $31:DriveStatusString:='31: no media in drive (IBM/MS INT 13 extensions)';
    $32:DriveStatusString:='32: incorrect drive type stored in CMOS (Compaq)';
    $AA:DriveStatusString:='AA: drive not ready (hard disk)';
    $B0:DriveStatusString:='B0: volume not locked in drive (INT 13 extensions)';
    $B1:DriveStatusString:='B1: volume locked in drive (INT 13 extensions)';
    $B2:DriveStatusString:='B2: volume not removable (INT 13 extensions)';
    $B3:DriveStatusString:='B3: volume in use (INT 13 extensions)';
    $B4:DriveStatusString:='B4: lock count exceeded (INT 13 extensions)';
    $B5:DriveStatusString:='B5: valid eject request failed (INT 13 extensions)';
    $BB:DriveStatusString:='BB: undefined error (hard disk)';
    $CC:DriveStatusString:='CC: write fault (hard disk)';
    $E0:DriveStatusString:='E0: status register error (hard disk)';
    $FF:DriveStatusString:='FF: sense operation failed (hard disk)';
  else
    DriveStatusString:='unknown error';
  end;
end;

Procedure PrintWelcome;
var
  foo:byte;
begin
  write('');
  PrintHeader('DISK2IMG','Dumps a diskette to a raw image file','1.11',20200409);
end;


function secstomins(secs:real):string;
var
  min,sec:string;
begin
  min:=inttostr(round(secs) div 60);
  sec:=inttostr(round(secs) mod 60);
  if sec[0]=#1
    then secstomins:=min+':0'+sec
    else secstomins:=min+':'+sec;
end;

Procedure DisplayStatus(s:string);
const
  statusline=23;
  blanks:string='                                                                               ';
begin
  {clear line}
  {fillchar(ptr(seg(vidp^),(statusline-1) * ScreenColumns * 2)^,ScreenColumns *2,0);}
  tfastwritexy(@blanks,vidp,0,statusline,emphcolor);
  tfastwritexy(@s,vidp,0,statusline,emphcolor);
end;

Procedure DisplayDriveStatus(_track,_head,_sector:byte;timeonly:boolean);

var
  percent:real;
  totalsecs:real;
  secselapsed:real;
  secsremaining:real;
  temps:string;

begin
  if not timeonly then begin
    tfastsetpos(0,0);
    tfastwriteln('   Drive: '+chr(65+CurDrive)+':');
    tfastwriteln('   Track: '+inttostr(_track)+' ');
    tfastwriteln('    Head: '+inttostr(_head)+' ');
    if _sector<>0
      then tfastwriteln('  Sector: '+inttostr(_sector)+'    ');
  end;

  secselapsed:=(TicksSinceMidnight-starttick)/tickHZ;
  if _sector=0 {doing a track dump}
    then percent:=(trackreadcount/totaltrackreads) * 100
    else percent:=(sectorcount/totalsectors) * 100;
  if (percent<>0)
    then totalsecs:=((TicksSinceMidnight-starttick)/tickHZ) / percent * 100
    else totalsecs:=1;
  secsremaining:=totalsecs-secselapsed;

  Str(percent:2:1,temps); temps:='Complete: '+temps+'%  ';
  tfastwritexy(@temps,vidp,6,9,normcolor);
  temps:='Time remaining: '+secstomins(secsremaining)+'  ';
  tfastwritexy(@temps,vidp,0,10,normcolor);
end;

Procedure SetupScreen;
begin
  {re-init BIOS mode we're already in to be kind to screen accelerator TSRs}
  asm
        mov     ax,0f00h
        int     10h                     {get existing mode}
        xor     ah,ah
        int     10h                     {(re)-set existing mode}
  end;
  {Set up fast textmode writing}
  tfastinit(80,25);
  tfPreventSnow:=true;
  tfastclrscr(vidp,$0f,0);
  tfastsetcolor(normcolor);
  tfastsetdest(vidp);
  tfastsetpos(0,0);
end;

Procedure DoneScreen;
begin
  tfastclrscr(vidp,$07,0);
  PositionCursor(0,23);
end;

procedure PositionCursor(x,y:byte); assembler;
asm
        {position "cursor"}
        mov     ah,02
        mov     bh,0
        mov     dh,y
        mov     dl,x
        int     10h
end;

Procedure DisplaySector(secbuf:pointer;line:byte);assembler;
asm
        push    ds
        xor     dx,dx
        mov     ax,(ScreenColumns * 2)
        mul     line
        les     di,vidp
        add     di,ax
        lds     si,secbuf
        add     di,(ScreenColumns * 2)-(secdisplaylinelength * 2) {location on screen to start}
        mov     ax,sectorsize
        mov     dl,secdisplaylinelength
        div     dl {ax contains sectorsize divided by sectordisplay line length which is number of lines to draw (8 usually)}
        mov     bh,al
        cld
        {call m6845_WaitVertRetrace}
        mov     bl,c_vertical_sync
        mov     dx,m6845_status
(*        mov     ah,sectcolor*) {fill ah with char display attribute}

@WDR: {wait during retrace, because we don't know where we are in the cycle}
        in      al,dx
        test    al,bl {if our bit is 1, then we're already in retrace, which means we missed it}
        jnz     @WDR  {jump if 1 (not 0) = keep looping as long as we're retracing}
@WDD: {wait for display to be over}
        in      al,dx
        test    al,bl
        jz      @WDD  {loop until we aren't drawing any more (ie. retracing)}

@displayloop:
        mov     cx,(secdisplaylinelength / 16) {unrolled}
@lineloop: {unrolled to 16 items}
        (*
        lodsb;  stosw 16 times - movsb ; inc di is slightly faster
        *)
        movsb; inc di; movsb; inc di; movsb; inc di; movsb; inc di;
        movsb; inc di; movsb; inc di; movsb; inc di; movsb; inc di;
        movsb; inc di; movsb; inc di; movsb; inc di; movsb; inc di;
        movsb; inc di; movsb; inc di; movsb; inc di; movsb; inc di;
        loop    @lineloop
        add     di,(ScreenColumns * 2)-(secdisplaylinelength * 2) {skip to next line}
        dec     bh {dec line count}
        jnz     @displayloop {if not done, do another line}
        pop     ds
end;

Procedure updateSalvageCounts;assembler;
{updates bit counter statistics}
asm
        mov     si,offset salvagedSector
        mov     di,offset salvageData
        mov     cx,sectorsize
        xor     bx,bx
        cld
@countbits:
        lodsb
        shl     al,1
        adc     word ptr [di],bx
        shl     al,1
        adc     word ptr [di+2],bx
        shl     al,1
        adc     word ptr [di+4],bx
        shl     al,1
        adc     word ptr [di+6],bx
        shl     al,1
        adc     word ptr [di+8],bx
        shl     al,1
        adc     word ptr [di+10],bx
        shl     al,1
        adc     word ptr [di+12],bx
        shl     al,1
        adc     word ptr [di+14],bx
        add     di,16
        loop    @countbits
end;

end.
